module mod_rt_raster_polygon
  use lib_const
  use lib_base
  use lib_time
  use lib_log
  use lib_array
  use lib_io
  use lib_math
  use def_common_params
  use def_common_types
  use mod_common_gs, only: &
    print_gs_raster, &
    print_gs_polygon, &
    !print_raster, &
    print_polygon
  use mod_common_area_raster_polygon, only: &
    set_modvars, &
    alloc_iarea, &
    free_iarea, &
    update_iarea_polygon, &
    update_iarea_sum, &
    fill_miss_iarea_sum, &
    calc_ifrac_sum, &
    update_rt1d
  use mod_common_rt, only: &
    calc_rt_im_nij_ulim, &
    init_rt1d, &
    reshape_rt1d, &
    free_rt1d_comps, &
    clear_rt_main, &
    output_rt_im
  use def_types
  implicit none
  private
  !-------------------------------------------------------------
  ! Public procedures
  !-------------------------------------------------------------
  public :: make_rt_raster_polygon
  !-------------------------------------------------------------
  logical, parameter :: debug_s = .false.
  logical, parameter :: debug_t = .false.
  logical, parameter :: debug = debug_s .or. debug_t
  integer(8), parameter :: sidx_debug = 0_8
  integer(8), parameter :: tidx_debug = 1301994_8
  !-------------------------------------------------------------
contains
!===============================================================
!
!===============================================================
subroutine make_rt_raster_polygon(s, t, rt, opt)
  implicit none
  type(gs_) , intent(inout), target :: s  ! raster
  type(gs_) , intent(inout), target :: t  ! polygon
  type(rt_) , intent(inout), target :: rt
  type(opt_), intent(in) :: opt

  type(gs_common_)      , pointer :: sgc, tgc
  type(gs_raster_)      , pointer :: sgr
  type(gs_polygon_)     , pointer :: tgp
  type(file_raster_in_) , pointer :: sfr
  type(file_polygon_in_), pointer :: tfp
  type(zone_latlon_)    , pointer :: szl
  type(zone_polygon_)   , pointer :: tzp
  type(grid_)           , pointer :: sg
  type(rt_main_)        , pointer :: rtm
  type(rt_im_zone_)     , pointer :: rtiz
  type(rt_vrf_)         , pointer :: rtv
  type(file_rt_vrf_)    , pointer :: fvrf
  type(rt1d_), allocatable :: rt1d(:)
  type(polygon_), pointer :: tp
  type(file_), pointer :: f

  integer(8) :: tij, tijs
  integer :: iFile_rtv
  real(8), pointer :: iarea(:,:)
  real(8), pointer :: iarea_sum(:,:), ifrac_sum(:,:)
  logical :: is_iarea_updated
  logical, parameter :: make_rt = .true.

  call echo(code%bgn, 'make_rt_raster_polygon')
  !-------------------------------------------------------------
  ! Set pointers
  !-------------------------------------------------------------
  sgc => s%cmn
  tgc => t%cmn

  if( sgc%gs_type /= gs_type_raster .or. &
      tgc%gs_type /= gs_type_polygon )then
    call eerr(str(msg_invalid_value())//&
            '\n  s%cmn%gs_type: '//str(sgc%gs_type)//&
            '\n  t%cmn%gs_type: '//str(tgc%gs_type))
  endif

  sgr => s%raster
  tgp => t%polygon

  sfr => sgr%f_raster_in
  tfp => tgp%f_polygon_in

  szl => sgr%zone(sgr%iZone)
  tzp => tgp%zone(tgp%iZone)

  sg  => sgr%grid

  if( sgc%is_source )then
    call print_gs_raster(&
           'Source', sgc%nam, &
           szl%typ, &
           szl%hi, szl%hf, szl%vi, szl%vf, &
           sgr%hi, sgr%hf, sgr%vi, sgr%vf, &
           szl%west, szl%east, szl%south, szl%north)
    call print_gs_polygon(&
           'Target', tgc%nam, &
           tzp%ijs, tzp%ije, tgp%ijs, tgp%ije)
  else
    call print_gs_polygon(&
           'Source', tgc%nam, &
           tzp%ijs, tzp%ije, tgp%ijs, tgp%ije)
    call print_gs_raster(&
           'Target', sgc%nam, &
           szl%typ, &
           szl%hi, szl%hf, szl%vi, szl%vf, &
           sgr%hi, sgr%hf, sgr%vi, sgr%vf, &
           szl%west, szl%east, szl%south, szl%north)
  endif

  rtm => rt%main
  rtiz => rt%im%zone(rt%im%iZone)
  !-------------------------------------------------------------
  ! Initialize
  !-------------------------------------------------------------
  nullify(iarea)
  nullify(iarea_sum)
  nullify(ifrac_sum)

  allocate(rt1d(tzp%mij))
  call init_rt1d(rt1d)

  call calc_rt_im_nij_ulim(rt%im%nij_ulim, opt%sys%memory_ulim)
  call edbg('rt%im%nij_ulim: '//str(rt%im%nij_ulim))

  rtiz%nij = 0_8
  rtm%nij = 0_8

  if( sgr%is_source )then
    rtv => rt%vrf_source
  else
    rtv => rt%vrf_target
  endif

  call set_modvars(sgr, tgp, make_rt)
  call alloc_iarea(iarea, iarea_sum, ifrac_sum)
  !-------------------------------------------------------------
  ! Make regridding table
  !-------------------------------------------------------------
  call echo(code%ent, 'Making regridding table')

  tijs = 1_8

  do tij = 1_8, tzp%mij
    tp => tgp%polygon(tij)
    if( tp%idx == tgp%idx_miss ) cycle

    !call edbg('tij: '//str(tij)//' idx: '//str(tp%idx))
    if( debug_t .and. tp%idx /= tidx_debug ) cycle
    !-----------------------------------------------------------
    ! Calc. intersection area
    !-----------------------------------------------------------
    call update_iarea_polygon(iarea, tp, is_iarea_updated, sgr%idxmap)
    call update_iarea_sum(iarea_sum, iarea, is_iarea_updated)
    if( is_iarea_updated )then
      call update_rt1d(rt1d(tij), iarea, sgr%idxmap, sgr%wgtmap, sg%idx, sg%idxarg)
    endif
    !-----------------------------------------------------------
    ! Output intermediates
    !-----------------------------------------------------------
    if( is_iarea_updated )then
      if( rt%im%nij_ulim > 0_8 )then
        if( rtm%nij > rt%im%nij_ulim )then
          call echo(code%ent, 'Outputting intermediates')
          call edbg('tij: '//str((/tijs,tij-1_8/),' ~ '))

          call reshape_rt1d(rt1d(tijs:tij-1_8), tgp%is_source, rtm, opt%earth)
          call output_rt_im(rtm, rt%im)
          call clear_rt_main(rtm)
          call free_rt1d_comps(rt1d(tijs:tij-1_8))
          tijs = tij

          call echo(code%ext)
        endif
      endif
    endif
    !-----------------------------------------------------------
    call inc(rtm%nij, rt1d(tij)%mij)
  enddo  ! tij/

  ! Output intermediates
  !-------------------------------------------------------------
  call echo(code%ent, 'Outputting intermediates')

  call reshape_rt1d(rt1d(tijs:tzp%mij), tgp%is_source, rtm, opt%earth)

  if( rt%im%nZones > 1 .or. rt%im%nij_max > 0_8 )then
    call output_rt_im(rtm, rt%im)
    call clear_rt_main(rtm)
    call free_rt1d_comps(rt1d(tijs:tzp%mij))
  endif

  call echo(code%ext)
  !-------------------------------------------------------------
  call echo(code%ext)
  !-------------------------------------------------------------
  ! Fill missing values for missing rasters
  !-------------------------------------------------------------
  call fill_miss_iarea_sum(iarea_sum, sgr%idxmap)
  !-------------------------------------------------------------
  ! Calc. $ifrac_sum
  !-------------------------------------------------------------
  call calc_ifrac_sum(ifrac_sum, iarea_sum, sgr%idxmap)
  !call calc_ifrac_sum(ifrac_sum, iarea_sum)
  !-------------------------------------------------------------
  ! Reverse
  !-------------------------------------------------------------
  if( .not. sgr%is_south_to_north )then
    call reverse(iarea_sum, 2)
    call reverse(ifrac_sum, 2)
  endif
  !-------------------------------------------------------------
  ! Output raster data for verification
  !-------------------------------------------------------------
  call echo(code%ent, 'Outputting raster data for verification')

  do iFile_rtv = 1, rtv%nFiles
    fvrf => rtv%f(iFile_rtv)

    f => fvrf%out_iarea_sum
    if( f%path /= '' )then
      call edbg('Writing iarea_sum '//str(fileinfo(f)))
      call write_binary(iarea_sum, f%path, f%dtype, f%rec, f%endian, &
                        sz=(/sgr%nh,sgr%nv/), lb=(/szl%xi,szl%yi/), fill=0.d0)
    endif

    f => fvrf%out_ifrac_sum
    if( f%path /= '' )then
      call edbg('Writing ifrac_sum '//str(fileinfo(f)))
      call write_binary(ifrac_sum, f%path, f%dtype, f%rec, f%endian, &
                        sz=(/sgr%nh,sgr%nv/), lb=(/szl%xi,szl%yi/), fill=0.d0)
    endif
  enddo  ! iFile_rtv/

  call echo(code%ext)
  !-------------------------------------------------------------
  !
  !-------------------------------------------------------------
  deallocate(rt1d)

  call free_iarea(iarea, iarea_sum, ifrac_sum)
  !-------------------------------------------------------------
  call echo(code%ret)
end subroutine make_rt_raster_polygon
!===============================================================
!
!===============================================================
end module mod_rt_raster_polygon
